﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace AitP
{
	public class AitPGame
	{
		public enum Role
		{
			Guard = 0,
			King = 1,
			Assassin = 2
		}

		public class SlotState
		{
			public Role Role { get; set; }
			public bool LedLynch { get; set; }
		}

		private static Random _random = new Random();
		private int _assassinLynchDay;
		private int _maxPlayers;
		private Log _logger;

		public delegate void Log(string message);

		public AitPGame(int maxPlayers, int assassinLynchDay, Log logger)
		{
			if (maxPlayers < 5) throw new ApplicationException("Too few players.");
			_maxPlayers = maxPlayers;
			_assassinLynchDay = assassinLynchDay;
			_logger = logger;
		}

		/// <summary>
		/// Returns true on town win, false on assassin win
		/// </summary>
		public bool Play()
		{
			List<SlotState> slots = Setup();
			int day = 1;

			while (true)
			{
				int assassinSlot = _GetAssassinSlot(slots);
				// Decide who is doing the lynch.  Once someone has led a lynch, they cannot lead other lynches,
				// so find someone else in that case.
				// If there's an Assassin lynch Day specified, the Assassin only lynches on or after that Day unless everyone else has already lynched.
				int lyncherSlot = _CheckSingleLyncher(slots);
				if (lyncherSlot == -1)
				{
					if (_assassinLynchDay < 1)
					{
						while (lyncherSlot == -1 || slots[lyncherSlot].LedLynch) lyncherSlot = _random.Next(slots.Count);
					}
					else
					{
						// Special handling: if this is before the Assassin's lynch Day, the assassin will not lynch.
						while (lyncherSlot == -1 || slots[lyncherSlot].LedLynch || (_assassinLynchDay > day && slots[lyncherSlot].Role == Role.Assassin)) lyncherSlot = _random.Next(slots.Count);
					}
				}
				
				// Determine who is being lynched.
				// The lyncher will never lynch themselves.  If the lyncher is not the Assassin, they will never lynch the King.
				int lyncheeSlot = -1;
				while (lyncheeSlot == -1 || lyncheeSlot == lyncherSlot || (slots[lyncheeSlot].Role == Role.King && slots[lyncherSlot].Role != Role.Assassin)) lyncheeSlot = _random.Next(slots.Count);

				// If the lynchee is the King, town loses.
				if (slots[lyncheeSlot].Role == Role.King)
				{
					if (_logger != null) _logger(string.Format("{0} lynched the King on Day {1}.  Assassin wins.", slots[lyncherSlot].Role, day));
					return false;
				}
				else if (slots[lyncheeSlot].Role == Role.Assassin)
				{
					// If the lynchee is the Assassin, he gets one shot to hit the King.
					int targetSlot = -1;
					while (targetSlot == -1 || targetSlot == lyncheeSlot) targetSlot = _random.Next(slots.Count);
					if (slots[targetSlot].Role == Role.King)
					{
						if (_logger != null) _logger(string.Format("{0} lynched the Assassin on Day {1}, but the Assassin shot the King.  Assassin wins.", slots[lyncherSlot].Role, day));
						return false;
					}
					else
					{
						if (_logger != null) _logger(string.Format("{0} lynched the Assassin on Day {1}, and the Assassin shot a Guard.  Town wins.", slots[lyncherSlot].Role, day));
						return true;
					}
				}
				else
				{
					// Otherwise, it was a guard.  Remove them and report.
					// if (_logger != null) _logger(string.Format("{0} lynched a Guard on Day {1}.", slots[lyncherSlot].Role, day));
					slots[lyncherSlot].LedLynch = true;
					slots.RemoveAt(lyncheeSlot);
					day++;
				}
			}
		}

		private List<SlotState> Setup()
		{
			List<SlotState> slots = new List<SlotState>(_maxPlayers);
			// King is random 0-based slot 
			int kingSlot = _random.Next(_maxPlayers);
			int assassinSlot = -1;
			// Assassin is random 0-based slot that is not the King slot
			while (assassinSlot == -1 || assassinSlot == kingSlot) assassinSlot = _random.Next(_maxPlayers);
			for (int ii = 0; ii < _maxPlayers; ii++)
			{
				if (ii == kingSlot)
				{
					slots.Add(new SlotState { Role = Role.King, LedLynch = false });
				}
				else if (ii == assassinSlot)
				{
					slots.Add(new SlotState { Role = Role.Assassin, LedLynch = false });
				}
				else
				{
					slots.Add(new SlotState { Role = Role.Guard, LedLynch = false });
				}
			}
			// if (_logger != null) _logger(string.Format("{0} player game: Assassin is slot #{1}, King is slot #{2}", _maxPlayers, assassinSlot, kingSlot));
			return slots;
		}

		private int _CheckSingleLyncher(List<SlotState> slots)
		{
			int notLynchedCount = 0;
			int notLynchedSlot = -1;
			for (int ii = 0; ii < slots.Count; ii++)
			{
				if (!slots[ii].LedLynch)
				{
					notLynchedCount++;
					notLynchedSlot = ii;
				}
			}
			if (notLynchedCount == 0)
			{
				// If we get here, everyone has led a lynch.  Clear the flag.
				for (int ii = 0; ii < slots.Count; ii++) slots[ii].LedLynch = false;
			}
			else if (notLynchedCount == 1)
			{
				return notLynchedSlot;
			}
			return -1;
		}
		private int _GetAssassinSlot(List<SlotState> slots)
		{
			for (int ii = 0; ii < slots.Count; ii++) if (slots[ii].Role == Role.Assassin) return ii;
			throw new ApplicationException("Could not find Assassin slot.");
		}
	}
}
